------------------------------------------------------------------------------
--  Thin Ada95 binding to OCI (Oracle Call Interface)                    --
--  Copyright (C) 2000-2002 Dmitriy Anisimkov.                              --
--  License agreement and authors contact information are in file oci.ads   --
------------------------------------------------------------------------------

--  $Id$

with OCI.Lib;
with OCI.Thread;
with Interfaces.C;
with Text_IO;
package body OCI.Thick.Debug_Allocation is

   use OCI.Lib;
   
   type Integer_Parameters is array (0 .. 16) of Integer;
   
   type Integer_Parameters_Access is access all Integer_Parameters;
   
--   pragma Convention (C, Integer_Parameters_Access);
   
   function Callback (
      ctxp : DVoid;
      hndlp : OCIHandle;
      htype : Ub4;
      fcode : Ub4;
      when_call : Ub1;
      returnCode : SWord;
      errnop : access Sb4;
      arglist : Integer_Parameters_Access) return SWord;
      
   pragma Convention (C, Callback);

   OCIHandleAlloc_FCode : constant := 2;
   OCIHandleFree_FCode : constant := 3;
   OCIDescriptorAlloc_FCode : constant := 4;
   OCIDescriptorFree_FCode  : constant := 5;
   
   HandleAlloc     : aliased constant String := "HandleAlloc";
   HandleFree      : aliased constant String := "HandleFree";
   DescriptorAlloc : aliased constant String := "DescriptorAlloc";
   DescriptorFree  : aliased constant String := "DescriptorFree";
   
   type String_Access is access constant String;
   
   Started : Boolean := False;
   
   subtype Func_Code is Integer
      range OCIHandleAlloc_FCode .. OCIDescriptorFree_FCode;
      
   subtype Handle_Type is Integer range OCI_HTYPE_FIRST .. OCI_HTYPE_LAST;
      
   subtype Call_Counter is Natural;
   
   Func_Names : array (Func_Code) of String_Access := 
     (OCIHandleAlloc_FCode => HandleAlloc'Access,
      OCIHandleFree_FCode => HandleFree'Access,
      OCIDescriptorAlloc_FCode => DescriptorAlloc'Access,
      OCIDescriptorFree_FCode => DescriptorFree'Access);
   
   Statistic : array (Func_Code) of Call_Counter := (others => 0);
                                 
   Error_Statistic : array (Func_Code) of Call_Counter := (others => 0);
   
   Handle_Allocated : array (Handle_Type) of Call_Counter := (others => 0);
   
   Handle_Freed : array (Handle_Type) of Call_Counter := (others => 0);
   
   protected Oper is
   
      procedure Increment (Item : in out Call_Counter);
   
   end Oper;
   
   protected body Oper is
   
      procedure Increment (Item : in out Call_Counter)
      is
      begin
         Item := Item + 1;
      end Increment;
   
   end Oper;
      
   function Callback (
      ctxp : DVoid;
      hndlp : OCIHandle;
      htype : Ub4;
      fcode : Ub4;
      when_call : Ub1;
      returnCode : SWord;
      errnop : access Sb4;
      arglist : Integer_Parameters_Access) return SWord
   is
      use Interfaces.C;
      Alloc_handle_type : integer renames arglist (2);
      Free_handle_type : integer renames arglist (1);
   begin
      if errnop.all = OCI_SUCCESS then
         Oper.Increment (Statistic (Func_Code (fcode)));
      else
         Oper.Increment (Error_Statistic (Func_Code (fcode)));
      end if;
--      Text_IO.Put_Line (Integer'Image (arglist (0)));
--      Text_IO.Put_Line (Integer'Image (arglist (1)));
--      Text_IO.Put_Line (Integer'Image (arglist (2)));
--      Text_IO.Put_Line (Integer'Image (arglist (3)));
--      Text_IO.Put_Line (Integer'Image (arglist (4)));
      if fcode = OCIHandleAlloc_FCode then
         if Alloc_handle_type in Handle_Type then
            Oper.Increment (Handle_Allocated (Alloc_handle_type));
         end if;
      elsif fcode = OCIHandleFree_FCode then
         if Free_handle_type in Handle_Type then
            Oper.Increment (Handle_Freed (Free_handle_type));
         end if;
      end if;
      return OCI_SUCCESS;
   end Callback;

   ----------------
   -- Print_Info --
   ----------------

   procedure Print_Info is
      use Text_IO;
   begin
      for I in Func_Code loop
         Put_Line (Func_Names (I).all &
            Integer'Image (Statistic (I)) &
            Integer'Image (Error_Statistic (I)));
      end loop;
      Put_Line ("Handles     " & Integer'Image (
         Statistic (OCIHAndleAlloc_FCode) - Statistic (OCIHAndleFree_FCode))
      );
      Put_Line ("Descriptors " & Integer'Image (
         Statistic (OCIDescriptorAlloc_FCode) - Statistic (OCIDescriptorFree_FCode))
      );
      for I in Handle_Type loop
        Put_Line (
            Integer'Image (I) & '.' &
            Integer'Image (Handle_Allocated (I)) & '-' &
            Integer'Image (Handle_Freed (I)) & '=' &
            Integer'Image (Handle_Allocated (I) - Handle_Freed (I)));
      end loop;
   end Print_Info;

   -----------
   -- Start --
   -----------

   procedure Start is
      Rc : SWord;
     
   begin
      if Started then
         raise Already_Started;
      else
         for fi in Func_Code loop
            Text_IO.Put_Line ("----------");
            Rc := OCIUserCallbackRegister (
                     hndlp => OCIHandle (Thread.Environment),
                     htype => OCI_HTYPE_ENV,
                     ehndlp => Thread.Error,
                     callback  => Callback'Address,
                     ctxp => Empty_Handle,
                     fcode => Ub4 (fi),
                     when_call => OCI_UCBTYPE_EXIT);
            Text_IO.Put_Line ("RC = " & SWord'Image(RC));
            Check_Error (rc);
         end loop;
      end if;
   end Start;

end OCI.Thick.Debug_Allocation;

