------------------------------------------------------------------------------
--  Ada95 Interface to Oracle RDBMS                                         --
--  Copyright (C) 2000-2001 Dmitriy Anisimkov.                              --
--  License text and authors contact information could be found in oci.ads  --
------------------------------------------------------------------------------
with OCI.Lib;
with OCI.Thread;
with Ada.Text_IO;
with Ada.Exceptions;
with System;

package body OCI.Thick.Strings is

   use Lib;

   use Interfaces.C;
   use Strings;
   use OCI.Thread;

   type Access_Size_T is access all size_t;

   type Buffer_Ptr is access all Char_Buffer;

   type A_A_Ub2 is access A_Ub2;
   type A_A_Sb2 is access A_Sb2;

   Type_Id : constant := OCI.Lib.SQLT_LNG;
   --  Type_Id : constant := OCI.Lib.SQLT_CHR;

   Indicator_Unchanged : constant := Sb2'First;

   String_Max_Size : constant := 32767;

   procedure Debug (Str : String);
   pragma Inline (Debug);

   function Callback_In_Bind
     (ictxp  : Value_Ptr;
      bindp  : OCIBind;
      iter   : Ub4;
      index  : Ub4;
      bufpp  : access Buffer_Ptr;
      alenp  : A_Ub4;
      piecep : A_Ub1;
      indp   : A_Sb2)
      return SWord;
   pragma Convention (C, Callback_In_Bind);
                             
   procedure Callback_Out
     (Value  : in     Value_Ptr;
      bufp   :    out Buffer_ptr;
      alenp  : in out Access_Size_T;
      piece  : in out Ub1;
      indp   :    out A_Sb2;
      Result :    out SWord);

   function Callback_Out_Bind
     (octxp   : Value_Ptr;
      bindp   : OCIBind;
      iter    : Ub4;
      index   : Ub4;
      bufpp   : access Buffer_ptr;
      alenpp  : access Access_Size_T;
      piecep  : A_Ub1;
      indpp   : A_A_Sb2;
      rcodepp : A_A_Ub2)
      return SWord;
   pragma Convention (C, Callback_Out_Bind);                                

   function Callback_Define
     (octxp   : Value_Ptr;
      defnp   : OCIDefine;
      iter    : Ub4;
      bufpp   : access Buffer_Ptr;
      alenpp  : access Access_Size_T;
      piecep  : A_Ub1;
      indpp   : A_A_Sb2;
      rcodepp : A_A_Ub2) return SWord;
   pragma Convention (C, Callback_Define);

   -----------
   -- Debug --
   -----------

   procedure Debug (Str : String) is
   begin
      --  pragma Debug (Ada.Text_IO.Put_Line (Str));
      null;
   end Debug;

   -----------------
   -- Print_Error --
   -----------------

   procedure Print_Error (Str : String) is
      use Ada.Text_IO;
   begin
      Put_Line (Current_Error, Str);
   end Print_Error;

   --------------------
   -- Set_Last_Value --
   --------------------

   procedure Set_Last_Value (Value : Value_Ptr) is
   begin
      Thread.Attributes.Reference.Last_Value := Value;
   end Set_Last_Value;
   
   --------------------
   -- Get_Last_Value --
   --------------------

   function Get_Last_Value return Value_Ptr is
   begin
      return Thread.Attributes.Reference.Last_Value;
   end Get_Last_Value;
   
   --------------------
   -- Get_Buffer_Ptr --
   --------------------

   function Get_Buffer_Ptr return Buffer_Ptr is
   begin
      return Thread.Attributes.Reference.Buffer'Access;
   end Get_Buffer_Ptr;

   -------------------------  
   -- Complete_Last_Value --
   -------------------------

   procedure Complete_Last_Value
   is
      Last_Value : Value_Ptr := Get_Last_Value;
      Buffer : Buffer_Ptr := Get_Buffer_Ptr;

      function String_Remain return String
      is
      begin
         return To_Ada (Buffer (0 .. Last_Value.alen - 1) & Nul, Trim_Nul => True);
      end String_Remain;
  
   begin
      if Last_Value = null then
         return;
      end if;
      pragma Assert (not Last_Value.Complete);

      Debug ("last value len " & size_t'Image (Last_Value.alen));

      if Last_Value.Indicator /= Null_Indicator then
         if Last_Value.First_Out_Piece then
            if Last_Value.Indicator = Indicator_Unchanged then
               Debug ("Unchanged indicator");
               Last_Value.Indicator := Last_Value.Prev_Indicator;
            else
               Last_Value.Value := To_Unbounded_String (String_Remain);
            end if;
         else
            Debug ("Not first last out");
            Append (Last_Value.Value, String_Remain);
         end if;
      end if;

      Last_Value.First_Out_Piece := False;
       
      Last_Value.alen := 0;

      Last_Value.Complete := True;

      -- Work around Oracle bug.
      if Last_Value.Indicator = Null_Indicator
         and Last_Value.rcode = 1405
      then
         Last_Value.rcode := OCI_SUCCESS;
      end if;

      if Last_Value.rcode /= OCI_SUCCESS then
         Ada.Exceptions.Raise_Exception
           (LIB_ERROR'Identity,
            "rcode = " & Ub2'Image (Last_Value.rcode));
      end if;

      Set_Last_Value (null);

   end Complete_Last_Value;
   
   ---------------
   -- Get_Value --
   ---------------

   function Get_Value (Var : Var_Type) return Unbounded_String is
   begin
      pragma Assert (Var.Complete);
      if OCI.Thick.Is_Null(Var) then
         raise NULL_VALUE;
      end if;
      return Var.Value;
   end Get_Value;
   
   function Get_Value (Var : Var_Type) return String is
   begin
      return To_String (Get_Value (Var));
   end Get_Value;  

   --------------
   -- Finalize --
   --------------

   procedure Finalize  (Object : in out Var_Type)
   is
      use type System.Address;
   begin
      if not Object.Complete then
         pragma Assert (Object'Address = Get_Last_Value.all'Address);
         Set_Last_Value (null);
         Object.Complete := true;
      end if;
      Debug ("final string");
      OCI.Thick.Finalize (Limited_Variable (Object));
   end Finalize;
      
   ---------------
   -- Set_Value --
   ---------------

   procedure Set_Value (Var : in out Var_Type; Value : Unbounded_String) is
   begin
      Var.Indicator := 0;
      Var.Value := Value;
      if not Var.Complete then
         Set_Last_Value (null);
         Var.First_Out_Piece := False;
         Var.Complete := true;
      end if;
   end Set_Value;
   
   procedure Set_Value (Var : in out Var_Type; Value : String) is
   begin
      Set_Value (Var, To_Unbounded_String (Value));
   end Set_Value;
      
   use type sb4;

   --  checks if remaining string, move it where need
   function Get_Buffer return Buffer_Ptr
   is
   begin
      Complete_Last_Value;
      return Get_Buffer_Ptr;
   end;
   
   ----------------------
   -- Callback_In_Bind --
   ----------------------

   function Callback_In_Bind
     (ictxp  : Value_Ptr;
      bindp  : OCIBind;
      iter   : Ub4;
      index  : Ub4;
      bufpp  : access Buffer_ptr;
      alenp  : A_Ub4;
      piecep : A_Ub1;
      indp   : A_Sb2) return SWord
   is
      Value : Value_Ptr renames ictxp;
      Next_Position : Positive := Value.Position + Char_Buffer'Length;
      Next_Value_Position : Positive;
   begin
      Debug
        ("callback in "
           & Ub4'Image (alenp.all)
           & Sb2'Image (Value.Indicator));

      indp.all := Value.Indicator;
        
      if Value.Value = Null_Unbounded_String 
      or Value.Indicator = Null_Indicator then
         alenp.all := 0;
         piecep.all := OCI_ONE_PIECE;
         indp.all := Null_Indicator;
         Debug ("null detected");
      else

         if Next_Position <= Length (Value.Value) then
            Next_Value_Position := Next_Position;
            if Value.Position = 1 then
               piecep.all := OCI_FIRST_PIECE;
               Debug ("OCI_FIRST_PIECE");
            else
               piecep.all := OCI_NEXT_PIECE;
               Debug ("OCI_NEXT_PIECE");
            end if;
         else
            Next_Value_Position := 1;
            if Value.Position = 1 then
               piecep.all := OCI_ONE_PIECE;
               Debug ("OCI_ONE_PIECE");
            else
               piecep.all := OCI_LAST_PIECE;
               Debug ("OCI_LAST_PIECE");
            end if;
         end if;

         bufpp.all := Get_Buffer;

         declare
            Part : String := Slice (Value.Value,
                  Value.Position, Natural'Min 
                        (Next_Position - 1, Length(Value.Value)));
            Moved : Size_T;
         begin
            Debug (Part);
            To_C (Part, bufpp.all.all, Moved, Append_Nul => False);
            alenp.all := Ub4 (Part'Length);
         end;
         
         Value.Position := Next_Value_Position;
         
      end if;
      Debug ("end in callback");
      return OCI_CONTINUE;
   exception
      when e: others =>
        Print_Error (Ada.Exceptions.Exception_Information(E));
        return OCI_Error;
   end Callback_In_Bind;
   
   ------------------
   -- Callback_Out --
   ------------------

   procedure Callback_Out
     (Value  : in     Value_Ptr;
      bufp   :    out Buffer_ptr;
      alenp  : in out Access_Size_T;
      piece  : in out Ub1;
      indp   :    out A_Sb2;
      Result :    out SWord)
   is

      function Part return String is
      begin
         return To_Ada (bufp.all (0 .. Value.alen - 1), False);
      end;

      use type A_Sb2;

   begin
      case piece is
      when OCI_ONE_PIECE | OCI_FIRST_PIECE =>
         Debug
           ("out OCI_one_first_PIECE " & Ub1'Image (piece));

         if Alenp /= null then
            Debug ("in out length " & Size_T'Image (alenp.all));
         end if;

         bufp := Get_Buffer;
         
         Bufp (Bufp'Last) := Nul; -- Debug.

         Value.alen := Char_Buffer'Length;
         alenp      := Value.alen'Access;
         indp       := Value.Indicator'Access;
         Value.Prev_Indicator := Value.Indicator;
         indp.all := Indicator_Unchanged;
         
         Value.Complete := False;
         Set_Last_Value (Value);
         Value.First_Out_Piece := True;

      when OCI_NEXT_PIECE =>
      
         Debug ("out OCI_NEXT_PIECE " & size_t'Image (Alenp.all));

         alenp := Value.alen'Access;
         indp  := Value.Indicator'Access;
         bufp  := Get_Buffer_Ptr;

         if Value.First_Out_Piece then
            Value.Value := To_Unbounded_String (Part);
            Value.First_Out_Piece := False;
         else
            Append (Value.Value, Part);
         end if;

         if Value.alen < Char_Buffer'Length then
            Value.Complete := True;
            Set_Last_Value (null);
            Piece := OCI_LAST_PIECE;
         else
            pragma Assert (not Value.Complete, "somebody got buffer");
            null;
         end if;

      when OCI_LAST_PIECE =>
         Debug ("out OCI_LAST_PIECE ");
         Append (Value.Value, Part);
         alenp.all := 0;
      when others => 
        Print_Error ("piece = " & Ub1'Image (piece));
        Result := OCI_Error;
     end case;  
     Result := OCI_CONTINUE;
   exception
      when e: others =>
         Print_Error
           ("In callback " & Ada.Exceptions.Exception_Information(E));
        Result := OCI_Error;
   end Callback_Out;

   -----------------------
   -- Callback_Out_Bind --
   -----------------------

   function Callback_Out_Bind
     (octxp   : Value_Ptr;
      bindp   : OCIBind;
      iter    : Ub4;
      index   : Ub4;
      bufpp   : access Buffer_ptr;
      alenpp  : access Access_Size_T;
      piecep  : A_Ub1;
      indpp   : A_A_Sb2;
      rcodepp : A_A_Ub2)
      return SWord
   is
      Result : SWord;
   begin
      rcodepp.all := octxp.rcode'Access;
      Callback_Out (octxp, bufpp.all, alenpp.all, piecep.all,
         indpp.all, Result);
      return Result;
   end Callback_Out_Bind;
   
   function Callback_Define
     (octxp   : Value_Ptr;
      defnp   : OCIDefine;
      iter    : Ub4;
      bufpp   : access Buffer_Ptr;
      alenpp  : access Access_Size_T;
      piecep  : A_Ub1;
      indpp   : A_A_Sb2;
      rcodepp : A_A_Ub2)
      return SWord
   is
      Result : SWord;
   begin
      Debug ("callback out index iter " & UB4'Image (iter));
      rcodepp.all := octxp.rcode'Access;
      Callback_Out (octxp, bufpp.all, alenpp.all, piecep.all,
         indpp.all, Result);
      return Result;
   end Callback_Define;

   procedure Bind_Dynamic (Value : in out Var_Type) is
      rc : SWord := OCIBindDynamic
        (bindp  => Value.Bind,
         errhp  => OCI.Thread.Error,
         ictxp  => Value'Address,
         icbfp  => Callback_In_Bind'Address,
         octxp  => Value'Address,
         ocbfp  => Callback_Out_Bind'Address);
   begin
      Check_Error(rc);
   end Bind_Dynamic;
  
   procedure Bind
     (Stmt  : in     Statement'Class;
      Value : in out Var_Type;
      Name  : in     String)
   is
       rc : sword := OCIBindByName (
                   stmtp => OCIStmt ( Handle (Stmt)),
                   bindpp => Value.Bind'Access,
                   errhp => OCI.Thread.Error,
                   placeholder => C.To_C(Name),
                   placeh_len => Name'Length,
                   valuep => System.Null_Address,
                   value_sz => String_Max_Size,
                   dty => Type_Id,
                   indp => Value.Indicator'Access,
                   mode => OCI_DATA_AT_EXEC);
   begin
     Check_Error (rc);
     Bind_Dynamic (Value);
   end Bind;

   procedure Bind
     (Stmt     : in     Statement'Class;
      Value    : in out Var_Type;
      Position : in     Positive)
   is
       rc : sword := OCIBindByPos
                  (stmtp => OCIStmt( Handle (Stmt)),
                   bindpp => Value.Bind'Access,
                   errhp => OCI.Thread.Error,
                   Position => ub4(Position),
                   valuep => System.Null_Address,
                   value_sz => String_Max_Size,
                   dty => Type_Id,
                   indp => Value.Indicator'Access,
                   mode => OCI_DATA_AT_EXEC);
   begin
     Check_Error(rc);
     Bind_Dynamic (Value);
   end Bind;

   procedure Define (Stmt : Statement'Class;
      Value : in out Var_Type;
      Position : Positive)
   is
      Rc : SWord := OCIDefineByPos
         (stmtp => OCIStmt( Handle (Stmt)),
          defnpp => Value.Define'Access,
          errhp => OCI.Thread.Error,
          position => ub4 (Position),
          value =>  Value'Address,
          value_sz => String_Max_Size,
          dty => Type_Id,
          indp => Value.Indicator'Access,
          mode => OCI_DYNAMIC_FETCH);
   begin
     Check_Error(rc);
     rc := OCIDefineDynamic
        (defnp  => Value.Define,
         errhp  => OCI.Thread.Error,
         octxp  => Value'Address,
         ocbfp  => Callback_Define'Address);
     Check_Error(rc);
   end Define;

end OCI.Thick.Strings;